/*
 * rc5.c - example implementation of RC5 Algorithm from Ronald L. Rivest
 *		   Dr.Dobbs Journal, January 1995 page 146,148
 *
 * basic crypt algorithm:
 *						A = A + S[0]
 *						B = B + S[1]
 *						for i=1 to r do
 *						   A = ((A^B)<<<B)+S[i*2]
 *						   B = ((B^A)<<<A)+S[i*2+1]
 *
 * the algorithm to generate S[] from the key was not published - so i used a random generator
 *
 * Date:	24.02.1995
 * Author:	Leopold Faschalek
 *
 * Adress:	Skidata Computer GmbH
 *			Untersbergstrasse 40
 *			A-5083 Gartenau
 *			Austria
 *
 * Fax:		+43-6246-888-7
 * CIS:		100010,3572
 *
 * RC5 and RSA-RC5 are trademarks of RSA Data Security Inc. Patent pending.
 *
 * this code is FREE - but in the USA you may have to contact RSA Data Security Inc.
 * the code implements an RC5-32/r/b where r=12 and b=8 in the example main()
 * this gives on my machine (Pentium 60MHz) 2MBytes/sec !!!!!!!!!!!!
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

typedef unsigned long word32;

#ifdef	_WIN32
#define ROL32(x,y)	_rotl(x,y)
#define ROR32(x,y)	_rotr(x,y)
#else
#define ROL32(x,y)	(((x)<<(y))|((x)>>(32-(y))))
#define ROR32(x,y)  (((x)>>(y))|((x)<<(32-(y))))
#endif

void rc5crypt(const word32 in[], word32 out[], const word32 S[], const int rounds)
{
	int i;
	word32 A,B;
#ifdef DECOMPOSED
	word32 A1,A2,A3,B1,B2,B3;
#endif
	
	A = in[0];
	B = in[1];
	
	A = A + S[0];
	B = B + S[1];
	for (i=1; i <= rounds; ++i){

		// original formula	from rc5
		// A = ((A ^ B) <<< B) + S[2*i]
		// B = ((B ^ A) <<< A) + S[2*i+1]
#ifdef DECOMPOSED
		// from this decomposed code you can compose the decrypt formula
		A1 = A ^ B;
		A2 = ROL32(A1,B&31);
		A3 = A2 + S[2*i];

		B1 = B ^ A3;
		B2 = ROL32(B1, A3&31);
		B3 = B2 + S[2*i+1];

		A = A3;	   
		B = B3;	   
#else
		A = ROL32(A^B,B)+S[i*2];  
		B = ROL32(B^A,A)+S[i*2+1];
#endif
	}	

	out[0] = A;
	out[1] = B;
}

void rc5decrypt(const word32 in[], word32 out[], const word32 S[], const int rounds)
{
	int i;
	word32 A,B;
#ifdef DECOMPOSED
	word32 A1,A2,A3,B1,B2,B3;
#endif
	
	A = in[0];
	B = in[1];
	
	for (i=rounds; i > 0; --i){

		// reverse the crypt operations
#ifdef DECOMPOSED
		A3 = A;
		B3 = B;
		
		B2 = B3 - S[2*i+1];
		B1 = ROR32(B2,A3&31);
		B  = B1 ^ A3;	

		A2 = A3 - S[2*i];
		A1 = ROR32(A2,B&31);
		A  = A1 ^ B;	
#else
		B = ROR32(B-S[i*2+1],A)^A;
		A = ROR32(A-S[i*2],B)^B;
#endif
	}
	A = A - S[0];
	B = B - S[1];

	out[0] = A;
	out[1] = B;
}

/////////////////////////////////////////////////////////////
//
// example random generator
//
static unsigned long int next = 1;

unsigned long my_rand(void)
{
	next = next * 1103515245L + 12345;
	return next;
}

void my_srand(unsigned long seed)
{
	next = seed;
}
//
/////////////////////////////////////////////////////////////

// there is no information how to exactly build the array S
// so i use a random generator with the key as seed
void rc5buildS(const word32 key[], const int keysize, word32 S[], const int Ssize)
{
	int i,j;

	memset(S, '\0', Ssize*sizeof(word32));
	for (i=0; i<keysize; ++i){
		my_srand(key[i]);
		for (j=0; j < Ssize; ++j){
			S[j] ^= my_rand();
		}
	}
}

#define NUMROUNDS	12
#define S_SIZE	2*(NUMROUNDS+1)

int main(int argc, char *argv[])
{
	word32  S[S_SIZE];
	word32  plain1[2] = { 0x55555555, 0xAAAAAAAA };
	word32  plain2[2] = { 0x55555556, 0xAAAAAAAA }; 
	word32  plain3[2] = { 0x0, 0x0 };
	word32  plain4[2] = { 0xFFFFFFFF, 0xFFFFFFFF };
	word32  cipher[2];
	word32	key[] = { 0x12345678, 0x9ABCDEF0 };

	rc5buildS( key, sizeof(key)/sizeof(key[0]), S, S_SIZE);

	printf("plain  %08X %08X\n",plain1[0],plain1[1]);
	rc5crypt( plain1, cipher, S, NUMROUNDS);
	printf("cipher %08X %08X\n",cipher[0],cipher[1]);
	rc5decrypt( cipher, plain1, S, NUMROUNDS);
	printf("plain  %08X %08X\n\n",plain1[0],plain1[1]);

	printf("plain  %08X %08X\n",plain2[0],plain2[1]);
	rc5crypt( plain2, cipher, S, NUMROUNDS);
	printf("cipher %08X %08X\n",cipher[0],cipher[1]);
	rc5decrypt( cipher, plain2, S, NUMROUNDS);
	printf("plain  %08X %08X\n\n",plain2[0],plain2[1]);

	printf("plain  %08X %08X\n",plain3[0],plain3[1]);
	rc5crypt( plain3, cipher, S, NUMROUNDS);
	printf("cipher %08X %08X\n",cipher[0],cipher[1]);
	rc5decrypt( cipher, plain3, S, NUMROUNDS);
	printf("plain  %08X %08X\n\n",plain3[0],plain3[1]);

	printf("plain  %08X %08X\n",plain4[0],plain4[1]);
	rc5crypt( plain4, cipher, S, NUMROUNDS);
	printf("cipher %08X %08X\n",cipher[0],cipher[1]);
	rc5decrypt( cipher, plain4, S, NUMROUNDS);
	printf("plain  %08X %08X\n\n",plain4[0],plain4[1]);

	// test crypt/encrypt and time it
	{
		int i;
		clock_t t1,t2;
		double seconds;
		const int loops = 1000000; // 1 million loops -> 8 MB Data
	
		rc5crypt( plain1, cipher, S, NUMROUNDS);
		t1 = clock();
		for (i=0; i<loops; i++)
			rc5crypt( cipher, cipher, S, NUMROUNDS);
		t2 = clock();
		seconds = ((double)(t2-t1))/CLOCKS_PER_SEC;
		printf("time for %d loops %7.3lf sec = %10.0lf Bytes/sec\n",loops,seconds,(loops*8)/seconds);

		for (i=0; i<loops; i++)
			rc5decrypt( cipher, cipher, S, NUMROUNDS);
		rc5decrypt( cipher, cipher, S, NUMROUNDS);

		if (plain1[0] == cipher[0] && plain1[1] == cipher[1])
			printf("RC5 is OK\n");
		else
			printf("RC5 Error !!!\n");
	}

	return 0;
}

/*
sample output on my COMPAQ DESPRO XL 560 (Pentium 60MHz) compiled with MSVC++ 2.0 under NT 3.5

plain  55555555 AAAAAAAA
cipher E69D80DF D378691B
plain  55555555 AAAAAAAA

plain  55555556 AAAAAAAA
cipher F72F8E94 A57B61D1
plain  55555556 AAAAAAAA

plain  00000000 00000000
cipher EE9D98C8 0BE73810
plain  00000000 00000000

plain  FFFFFFFF FFFFFFFF
cipher 779F3767 62174915
plain  FFFFFFFF FFFFFFFF

time for 1000000 loops   3.685 sec =    2170963 Bytes/sec
RC5 is OK

*/

